import os
from configparser import ConfigParser

from flask import Flask, g, redirect, url_for
from flask_cors import CORS

from twitter_v2_facade import twitter_app as twitter_v2
from twitter_archive_facade import twitter_app as twitter_archive


def import_env ():
    cp = ConfigParser()
    if os.path.exists('.env'):
        with open('.env') as stream:
            cp.read_string('[default]\n' + stream.read())
            os.environ.update(dict(cp['default']))

if __name__ == '__main__':
    import_env()

    PORT = int(os.environ.get('PORT', 5000))
    HOST = os.environ.get('HOST', '127.0.0.1')
    
    archive_enabled = os.environ.get('ARCHIVE_TWEETS_PATH') and True
    glitch_enabled = os.environ.get('PROJECT_DOMAIN') and True
    
    api = Flask(__name__, static_url_path='')
    
    

    @api.before_request
    def add_config ():
        g.glitch_enabled = glitch_enabled
        
        if glitch_enabled:
            g.app_url = 'https://{}.glitch.me'.format( os.environ.get('PROJECT_DOMAIN') )
        else:
            g.app_url = 'http://{}:{}'.format('localhost', PORT)
        

    
    @api.context_processor
    def inject_config ():
        return {'archive_enabled': archive_enabled}
    
    api.secret_key = os.environ.get('FLASK_SECRET')
    
    api.config['TEMPLATES_AUTO_RELOAD'] = True

    api.register_blueprint(twitter_v2, url_prefix='/twitter')
    
    if archive_enabled:
        api.register_blueprint(twitter_archive, url_prefix='/twitter-archive')
    
    CORS(api)
    
    @api.route('/')
    def index ():
        return redirect(url_for('.twitter_v2_facade.get_login_html'))
    
    
    api.run(port=PORT, host=HOST)